__author__ = 'Lijun Yu'

from .runtime import Options
from .stage import ReorderStage, Stage
from .system import Job, Result, System
from .task import Task
from .utils import get_logger, progressbar, process_map



from itertools import zip_longest
from typing import Iterable, List, Optional

from .runtime import mp
from .stage import Stage
from .task import ControlCommand, ControlTask, Task
from .worker import WorkerGroup


class Pipeline(object):

    '''
    A pipeline consists of multiple stages that executes sequentially.
    '''

    def __init__(self, stages: List[Stage]):
        self.stages = stages


class SyncPipeline(Pipeline):

    '''
    A simple synchronized pipeline for debug. 
    All stages run in the main process.
    '''

    def __init__(self,  stages: List[Stage], queue_size: int = 16):
        super(SyncPipeline, self).__init__(stages)

    def start(self, timeout: Optional[int] = 1):
        for stage in self.stages:
            stage.init()
        self.reset()

    def run_task(self, task: Task, current_stage: int = 0) -> Iterable[Task]:
        if current_stage == len(self.stages):
            yield task
            return
        current_result = self.stages[current_stage].run(task)
        if isinstance(current_result, Task):
            yield from self.run_task(current_result, current_stage + 1)
        else:
            for r in current_result:
                yield from self.run_task(r, current_stage + 1)

    def run(self, tasks: List[Task]) -> Iterable[Task]:
        for task in tasks:
            yield from self.run_task(task)

    def reset(self, timeout: Optional[int] = 1):
        for stage in self.stages:
            stage.reset()

    def end(self, timeout: Optional[int] = 1):
        return

    def join(self, timeout: Optional[int] = 1):
        return

    def terminate(self, timeout: Optional[int] = 1):
        return


class AsyncPipeline(Pipeline):

    '''
    Asynchronized pipeline powered by multiprocessing.
    '''

    def __init__(self, stages: List[Stage], queue_size: Optional[int] = 16):
        super(AsyncPipeline, self).__init__(stages)
        self.manager = mp.Manager()
        self.job_queue = self.manager.Queue()
        self.worker_groups = []
        job_queue = self.job_queue
        for stage, next_stage in zip_longest(stages, stages[1:]):
            group = WorkerGroup(
                stage, job_queue, self.manager, queue_size, next_stage)
            self.worker_groups.append(group)
            job_queue = group.result_queue
        self.result_queue = job_queue
        self.resetting = False
        self.terminated = False

    def start(self, timeout: Optional[int] = 1):
        for group in self.worker_groups:
            group.start()
        self.reset(timeout)

    def reset(self, timeout: Optional[int] = 1):
        assert not self.resetting, 'Cannot reset when reset is in progress'
        reset_task = ControlTask(ControlCommand.Reset)
        for _ in range(self.stages[0].num_worker):
            self.job_queue.put(reset_task, timeout=timeout)
        self.resetting = True

    def wait(self, timeout: Optional[int] = None):
        while True:
            result = self.result_queue.get(timeout=timeout)
            if isinstance(result, ControlTask):
                if result.command == ControlCommand.Reset:
                    self.resetting = False
                break
            else:
                yield result

    def end(self, timeout: Optional[int] = 1):
        if self.terminated:
            return
        end_task = ControlTask(ControlCommand.End)
        for _ in range(self.stages[0].num_worker):
            self.job_queue.put(end_task, timeout=timeout)
        for _ in self.wait(timeout):
            pass

    def join(self, timeout: Optional[int] = 1):
        for group in self.worker_groups:
            group.join(timeout)

    def terminate(self, timeout: Optional[int] = 3):
        for group in self.worker_groups:
            group.terminate(timeout)
        self.terminated = True



import time
from collections import namedtuple
from enum import IntEnum
from typing import Any, Optional

from .utils import get_logger

ControlCommand = IntEnum('ControlCommand', ['End', 'Reset'])


class ControlTask(object):
    '''
    Pipeline control task.
    '''

    def __init__(self, command: ControlCommand):
        self.command = command

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.command.name)


TaskLog = namedtuple('TaskLog', [
    'stage', 'duration', 'start_time', 'end_time', 'success'])


class Task(object):
    '''
    Regular excutable task.
    '''

    def __init__(self, content: Optional[Any] = None,
                 meta: Optional[dict] = None,
                 *, parent_task: Optional[Any] = None):
        self.create_time = time.time()
        self.content = content
        self._build_meta(meta, parent_task)
        self.logs = parent_task.logs.copy() if parent_task is not None else []
        self.success = True

    def _build_meta(self, meta, parent_task):
        self.meta = {}
        if parent_task is not None:
            self.meta.update(parent_task.meta)
        if meta is not None:
            self.meta.update(meta)

    def start(self, stage: Any):
        self.start_time = time.time()
        self.current_stage = str(stage)
        return self

    def finish(self, content: Optional[Any] = None, success: bool = True):
        self.finish_time = time.time()
        if content is not None:
            self.content = content
        duration = self.finish_time - self.start_time
        log = TaskLog(
            self.current_stage, duration, self.start_time, self.finish_time,
            success)
        self.logs.append(log)
        return self

    def fail(self):
        try:
            self.finish(success=False)
        except:
            get_logger(__name__).exception(
                'Exception ignored for failed task: %s', self)
        self.success = False

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, repr(self.meta))



import faulthandler
import os
import resource
import warnings

try:
    '''
    Prefer Pytorch's multiprocessing module for optimized sharing of Tensor
    '''
    import torch.multiprocessing as mp
    hard_limit = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
    resource.setrlimit(resource.RLIMIT_NOFILE, (hard_limit, hard_limit))
    if hard_limit < 51200:
        warnings.warn(
            'Ulimit of open files (%d) is too small. If errors of unable to '
            'open files occur, try add the following in your start '
            'script: `torch.mp.set_sharing_strategy("file_system")`' % (
                hard_limit))
except ImportError:
    import multiprocessing as mp

mp = mp.get_context('spawn')
faulthandler.enable()

QUEUE_EXCEPTIONS = (BrokenPipeError, ConnectionResetError, EOFError,
                    FileNotFoundError)


class Options(object):

    raise_exception = False
    single_sync_pipeline = False
    no_progress_bar = False
    print_debug_log = False
    log_file = None

    def __init__(self):
        for option in os.environ.get('PYTURBO_OPTIONS', '').split():
            fields = option.split('=')
            if len(fields) > 2:
                warnings.warn('Invalid option format: %s' % (option))
                continue
            name = fields[0]
            value = fields[1] if len(fields) == 2 else True
            if not hasattr(self, name):
                warnings.warn('Unrecognized option: %s' % (option))
                continue
            setattr(self, name, value)

    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        if name == 'log_file' and value is not None:
            faulthandler.enable(open(value))

    def __str__(self):
        return ', '.join(
            ['%s=%s' % (k, v) for k, v in self.__dict__.items()
             if not k.startswith('_') and v]) or 'None'

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, str(self))


Options = Options()



from typing import Optional

from .runtime import QUEUE_EXCEPTIONS, mp
from .stage import Stage
from .task import ControlCommand, ControlTask, Task


class WorkerGroup(object):

    '''
    A group of processes for one stage.
    '''

    def __init__(self, stage: Stage, job_queue: mp.Queue, manager: mp.Manager,
                 result_queue_size: int, next_stage: Optional[Stage] = None):
        self.stage = stage
        self.job_queue = job_queue
        self.manager = manager
        self.result_queue = self.manager.Queue(result_queue_size)
        self.control_barrier = self.manager.Barrier(stage.num_worker)
        self.processes = []
        next_num_worker = next_stage.num_worker if next_stage is not None else 1
        for worker_id in range(stage.num_worker):
            process = Worker(
                stage, worker_id, job_queue, self.result_queue,
                self.control_barrier, next_num_worker)
            self.processes.append(process)

    def start(self):
        for process in self.processes:
            process.start()

    def join(self, timeout: Optional[int] = 1):
        for process in self.processes:
            if process._closed:
                continue
            process.join(timeout)

    def terminate(self, timeout: Optional[int] = 3):
        for process in self.processes:
            if process._closed:
                continue
            process.terminate()
        self.join(timeout)


class Worker(mp.Process):

    '''
    One worker process.
    '''

    def __init__(self, stage: Stage, worker_id: int,
                 job_queue: mp.Queue, result_queue: mp.Queue,
                 control_barrier: mp.Barrier, next_num_worker: int):
        name = '%s-%d' % (stage.__class__.__name__, worker_id)
        super(Worker, self).__init__(name=name)
        self.stage = stage
        self.worker_id = worker_id
        self.job_queue = job_queue
        self.result_queue = result_queue
        self.control_barrier = control_barrier
        self.next_num_worker = next_num_worker

    def control(self, task: Task):
        if task.command == ControlCommand.Reset:
            self.stage.reset()
        pass_id = self.control_barrier.wait()
        if pass_id == 0:
            for _ in range(self.next_num_worker):
                self.result_queue.put(task)
        if task.command == ControlCommand.End:
            return True

    def run(self):
        self.stage.init(self.worker_id)
        while True:
            try:
                try:
                    task = self.job_queue.get()
                except QUEUE_EXCEPTIONS:
                    break
                if isinstance(task, ControlTask):
                    if self.control(task):
                        return
                    continue
                result = self.stage.run(task)
                if result is None:
                    continue
                if isinstance(result, Task):
                    result = [result]
                for r in result:
                    try:
                        self.result_queue.put(r)
                    except QUEUE_EXCEPTIONS:
                        break
            except (KeyboardInterrupt, GeneratorExit):
                break



import logging
import multiprocessing
from typing import Iterable, Optional

import psutil
from tqdm.autonotebook import tqdm

from .runtime import Options


def progressbar(iterable: Optional[Iterable] = None, desc: Optional[str] = None,
                total: Optional[int] = None, *, silent: bool = False,
                **kwargs):
    if not silent:
        identity = multiprocessing.current_process()._identity
        if 'position' not in kwargs and len(identity):
            kwargs['position'] = identity[0]
        return tqdm(iterable, desc, total, dynamic_ncols=True, **kwargs)
    return iterable


def get_logger(name: str, level: Optional[int] = None,
               log_file: Optional[str] = None):
    if level is None:
        level = logging.INFO if not Options.print_debug_log else logging.DEBUG
    if log_file is None:
        log_file = Options.log_file
    logger = logging.getLogger(name)
    logger.setLevel(level)
    if len(logger.handlers) > 0:
        return logger
    formatter = logging.Formatter(
        '%(asctime)s %(name)s %(levelname)s %(message)s')
    handlers = [logging.StreamHandler()]
    if log_file is not None:
        handlers.append(logging.FileHandler(log_file))
    for handler in handlers:
        handler.setLevel(level)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    return logger


def process_map(worker, jobs, num_worker=None):
    if num_worker is None:
        num_worker = len(psutil.Process().cpu_affinity())
    with multiprocessing.Pool(num_worker) as pool:
        yield from progressbar(
            pool.imap_unordered(worker, jobs), 'Jobs', len(jobs))



from functools import partial
from queue import SimpleQueue
from threading import Event, Thread
from typing import Iterable, List, Optional

from .pipeline import AsyncPipeline, SyncPipeline
from .resource import Resources
from .runtime import Options
from .stage import Stage
from .task import Task
from .utils import get_logger, progressbar


class Job(object):

    '''
    A job for the system. A job typically contains a task for the first stage 
    and some meta data.
    '''

    def __init__(self, name: str, task: Task, length: Optional[int] = None,
                 max_retry: int = 1):
        self.name = name
        self.task = task
        self.length = length
        self.retry = max_retry
        self.success = False
        self.results = None

    def finish(self, pipeline_id: int):
        self.pipeline_id = pipeline_id
        self.success = True

    def __repr__(self):
        return '%s(%s, len=%d, retry=%d)' % (
            self.__class__.__name__, self.name, self.length, self.retry)


class Result(object):

    '''
    A result container for a job.
    '''

    def __init__(self):
        self.results = []
        self.failed_tasks = []
        self.task_count = 0

    def __len__(self):
        return self.task_count

    def collect(self, task):
        '''
        Collect result from a task, called in System.monit_pipeline.
        '''
        self.task_count += 1
        if task.success:
            self.results.append(task.content)
        else:
            self.failed_tasks.append(task)

    def merge(self, job):
        '''
        Merge and post-process collected results, called in System.wait_job.
        '''
        return


class System(object):

    '''
    A system consists of multiple peer pipelines.
    '''

    def __init__(self, *, pipeline_queue_size: int = 16,
                 resource_scan_args={}, **custom_args):
        self.pipeline_queue_size = pipeline_queue_size
        self.debug_mode = Options.single_sync_pipeline
        self.resources = Resources(**resource_scan_args)
        self.num_pipeline = self.get_num_pipeline(
            self.resources, **custom_args)
        if self.debug_mode:
            self.num_pipeline = 1
            self.pipeline_cls = SyncPipeline
        else:
            self.pipeline_cls = AsyncPipeline
        self.logger = get_logger(repr(self))
        self.logger.info(repr(self.resources))
        self.logger.info(repr(Options))
        if self.debug_mode:
            self.logger.info(
                'Debug mode, one SyncPipeline')
        else:
            self.logger.info(
                'Production mode, %d AsyncPipelines', self.num_pipeline)
        self.job_queue = SimpleQueue()
        self.result_queue = SimpleQueue()
        self.ending = Event()
        self.job_count, self.result_count = 0, 0
        self.build()

    def get_num_pipeline(self, resources: Resources) -> int:
        return 1

    def get_stages(self, resources: Resources) -> List[Stage]:
        '''
        Define the stages in a pipeline with given resources.
        '''
        raise NotImplementedError

    def collect_results(self, job: Job, results_gen: Iterable[Task], *,
                        result_cls: type = Result):
        '''
        Define how to extract final results from output tasks.
        '''
        job.results = result_cls()
        for task in results_gen:
            try:
                job.results.collect(task)
            except:
                self.logger.exception(
                    'Failed to collect results from %s', task)

    def monit_pipeline(self, pipeline_id: int, *,
                       task_timeout: Optional[int] = None,
                       queue_timeout: Optional[int] = None):
        logger = get_logger(
            '%s(%d)' % (self.pipeline_cls.__name__, pipeline_id))
        job, job_ok = None, True
        pipeline_ok = True
        try:
            while True:
                job = self.job_queue.get()
                if job is None:
                    return
                job_ok = False
                pipeline = self.pipelines[pipeline_id]
                if not pipeline_ok:
                    logger.info('Restarting')
                    pipeline = self.pipeline_fns[pipeline_id]()
                    pipeline.start(queue_timeout)
                    [*pipeline.wait(task_timeout)]
                    self.pipelines[pipeline_id] = pipeline
                    logger.info('Restarted')
                pipeline_ok = False
                try:
                    if Options.no_progress_bar:
                        logger.info('Processing: %s', job)
                    if not self.debug_mode:
                        pipeline.job_queue.put(job.task, timeout=queue_timeout)
                        pipeline.reset(queue_timeout)
                        results_gen = pipeline.wait(task_timeout)
                    else:
                        results_gen = pipeline.run_task(job.task)
                    results_gen = progressbar(
                        results_gen, desc=' Pipeline %d (%s)' % (
                            pipeline_id, job.name),
                        total=job.length, position=pipeline_id, leave=False,
                        silent=Options.no_progress_bar)
                    self.collect_results(job, results_gen)
                    if self.debug_mode:
                        pipeline.reset(queue_timeout)
                    pipeline_ok = True
                    job.finish(pipeline_id)
                    if Options.no_progress_bar:
                        logger.info('Processed: %s', job)
                except Exception as e:
                    if self.ending.is_set():
                        return
                    logger.exception('Failed: %s', job)
                    if Options.raise_exception:
                        raise e
                    if not pipeline_ok:
                        logger.info('Terminating')
                        pipeline.terminate(queue_timeout * 3)
                finally:
                    self.result_queue.put(job, timeout=queue_timeout)
                    job_ok = True
        except Exception as e:
            logger.exception('Dead')
            if not job_ok:
                job.retry += 1
                self.result_queue.put(job, timeout=queue_timeout)
                job_ok = True
            self.result_queue.put(None, timeout=queue_timeout)
            if Options.raise_exception:
                raise e

    def build(self):
        self.pipeline_fns, self.pipelines = [], []
        for pipeline_i, resources in enumerate(
                self.resources.split(self.num_pipeline)):
            stages = self.get_stages(resources)
            self.logger.info('Building pipeline %d: \n\t%s', pipeline_i,
                             '\n\t'.join([repr(s) for s in stages]))
            pipeline_fn = partial(
                self.pipeline_cls, stages, self.pipeline_queue_size)
            self.pipeline_fns.append(pipeline_fn)
            pipeline = pipeline_fn()
            self.pipelines.append(pipeline)
        self.progressbar = None

    def start(self, *, task_timeout: Optional[int] = 600,
              queue_timeout: Optional[int] = 1):
        self.logger.info('Starting')
        for pipeline in self.pipelines:
            pipeline.start(queue_timeout)
        if not self.debug_mode:
            self.monitor_threads = []
            for i in range(self.num_pipeline):
                thread = Thread(
                    target=self.monit_pipeline, args=(i,),
                    kwargs=dict(task_timeout=task_timeout,
                                queue_timeout=queue_timeout), daemon=True)
                thread.start()
                self.monitor_threads.append(thread)
            for pipeline in self.pipelines:
                [*pipeline.wait(task_timeout)]
        self.logger.info('Started')
        if not Options.no_progress_bar:
            self.progressbar = progressbar(
                total=self.job_count, desc='Jobs', position=self.num_pipeline)

    def add_job(self, job: Job, *, timeout: Optional[int] = 1):
        self.job_queue.put(job, timeout=timeout)
        self.job_count += 1
        if not Options.no_progress_bar:
            self.progressbar.total = self.job_count
            self.progressbar.refresh()
        if self.debug_mode:
            self.job_queue.put(None, timeout=timeout)

    def add_jobs(self, jobs: List[Job], *, timeout: Optional[int] = 1):
        for job in jobs:
            self.add_job(job, timeout=timeout)

    def wait_job(self, *, job_timeout: Optional[int] = 3600,
                 queue_timeout: Optional[int] = 1):
        job = None
        if self.debug_mode:
            self.monit_pipeline(0)
        while self.num_pipeline > 0:
            job = self.result_queue.get(timeout=job_timeout)
            if job is None:
                self.num_pipeline -= 1
                continue
            if not job.success:
                if job.retry > 0:
                    job.retry -= 1
                    self.logger.info('Enqueue for retry: %s', job)
                    self.job_queue.put(job, timeout=queue_timeout)
                    continue
                if job.results is not None:
                    self.logger.warn(
                        'Partial results collected / total: %d / %d from %s',
                        len(job.results), job.length, job)
                else:
                    self.logger.warn('Failed: %s', job)
            break
        else:
            raise ValueError('All pipelines dead')
        if job.results is not None:
            try:
                job.results.merge(job)
            except:
                self.logger.exception('Failed to merge results: %s', job)
        self.result_count += 1
        if not Options.no_progress_bar:
            self.progressbar.update()
        else:
            self.logger.info('Jobs processed / total: %d / %d' % (
                self.result_count, self.job_count))
        return job

    def wait_jobs(self, num_jobs: int, *, job_timeout: Optional[int] = 3600,
                  queue_timeout: Optional[int] = 1):
        for _ in range(num_jobs):
            yield self.wait_job(
                job_timeout=job_timeout, queue_timeout=queue_timeout)

    def end(self, *, timeout: Optional[int] = 1):
        self.logger.info('Ending')
        assert self.job_queue.empty(), 'Ending with unprocessed jobs'
        assert self.result_queue.empty(), 'Ending with retrieved results'
        self.ending.set()
        for pipeline in self.pipelines:
            pipeline.end(timeout)
        for pipeline in self.pipelines:
            pipeline.join(timeout)
        if not self.debug_mode:
            for _ in range(self.num_pipeline):
                self.job_queue.put(None, timeout=timeout)
            for thread in self.monitor_threads:
                thread.join(timeout)
        if not Options.no_progress_bar:
            self.progressbar.close()
        self.logger.info('Ended')

    def terminate(self, *, timeout: Optional[int] = 3):
        self.logger.exception('Terminating')
        self.ending.set()
        for pipeline in self.pipelines:
            pipeline.terminate(timeout)
        if not self.debug_mode:
            for _ in range(self.num_pipeline):
                self.job_queue.put(None, timeout=timeout)
            for thread in self.monitor_threads:
                thread.join(timeout)
        if not Options.no_progress_bar:
            self.progressbar.close()
        self.logger.error('Terminated')

    def __repr__(self):
        return '%s(%s)' % (
            self.__class__.__name__,
            'debug' if self.debug_mode else 'x%d' % (self.num_pipeline))



import setuptools

with open('README.md') as f:
    long_description = f.read()

with open('requirements.txt') as f:
    requirements = f.readlines()

setuptools.setup(
    name='py-turbo',
    version='0.6.3',
    author='Lijun Yu',
    author_email='lijun@lj-y.com',
    description='A pipeline system for efficient execution.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/CMU-INF-DIVA/pyturbo',
    license='GPL',
    packages=setuptools.find_packages(),
    install_requires=requirements,
    python_requires='>=3.7',
)



from collections import deque
from typing import Iterable, List, Union

from .resource import Resources
from .runtime import Options
from .task import Task
from .utils import get_logger


class Stage(object):

    '''
    One stage in a pipeline. A stage can have multiple peer worker processes.
    Worker number is desided based on how available resources are allocated.
    '''

    def __init__(self, resources: Resources, **custom_args):
        self.resources = resources
        self.logger = get_logger(repr(self))
        self.resource_allocation = self.allocate_resource(
            resources, **custom_args)
        self.num_worker = len(self.resource_allocation)

    def allocate_resource(self, resources: Resources,
                          **custom_args) -> List[Resources]:
        '''
        Allocate resources to each worker, the length of return value indicates 
        number of workers. Called during object initialization with the custom 
        keyword arguments.
        '''
        if len(custom_args) > 0:
            self.logger.warn('Unprocessed custom_args: %s', custom_args)
        return [resources]

    def reset(self):
        '''
        Reset a worker process. 
        Executed during initialization and at reset command.
        '''
        self.logger.debug('Reset.')

    def process(self, task: Task) -> Union[Task, Iterable[Task]]:
        '''
        Process function for each worker process. 
        Returns one or a series of downstream tasks.
        '''
        raise NotImplementedError

    def init(self, worker_id: int = 0):
        self.worker_id = worker_id
        self.current_resource = self.resource_allocation[worker_id]
        name = '%s[%d]%s' % (self.__class__.__name__, worker_id,
                             '@' + str(self.current_resource)
                             if self.current_resource is not None else '')
        self.logger = get_logger(name)

    def run(self, task: Task) -> Union[Task, Iterable[Task]]:
        assert isinstance(task, Task), '%s is not a subclass of %s' % (
            type(task), Task)
        try:
            if not task.success:
                self.logger.warn('Bypassing: %s', task)
                yield task
                return
            self.logger.debug('Processing: %s', task)
            result = self.process(task)
            if isinstance(result, Task):
                yield result
            else:
                for res in result:
                    yield res
            self.logger.debug('Processed: %s', task)
        except Exception as e:
            task.fail()
            self.logger.exception('Failed: %s', task)
            if Options.raise_exception:
                raise e
            else:
                yield task

    def __repr__(self):
        if hasattr(self, 'num_worker'):
            return '%s(x%d)@(%s)' % (self.__class__.__name__, self.num_worker,
                                     self.resources)
        return self.__class__.__name__


class ReorderStage(Stage):

    '''
    A special stage with only one worker. 
    Tasks are processed according to a pre-defined order.
    '''

    def __init__(self, resources: Resources, reorder_buffer_size: int = 16,
                 **custom_args):
        super(ReorderStage, self).__init__(resources, **custom_args)
        assert self.num_worker == 1, 'ReorderStage must have only 1 worker.'
        self.reorder_buffer_size = reorder_buffer_size
        self.reorder_buffer_size_record = reorder_buffer_size

    def get_sequence_id(self, task: Task) -> int:
        '''
        Return the order of each task. Start from 0.
        '''
        raise NotImplementedError

    def reset(self):
        super(ReorderStage, self).reset()
        self.reorder_buffer = deque()
        self.next_id = 0

    def push(self, task: Task):
        try:
            offset = self.get_sequence_id(task) - self.next_id
            if offset >= len(self.reorder_buffer):
                self.reorder_buffer.extend([None] * (
                    offset - len(self.reorder_buffer) + 1))
            self.reorder_buffer[offset] = task
        except:
            if len(self.reorder_buffer) > 0 and self.reorder_buffer[0] is None:
                self.reorder_buffer[0] = task
            else:
                self.reorder_buffer.insert(0, task)
        if len(self.reorder_buffer) > self.reorder_buffer_size_record:
            self.reorder_buffer_size_record = len(self.reorder_buffer)
            self.logger.warn(
                'Reorder buffer size %d exceeds limit of %d',
                len(self.reorder_buffer), self.reorder_buffer_size)

    def iter_pop(self) -> Iterable[Task]:
        while len(self.reorder_buffer) > 0:
            if self.reorder_buffer[0] is not None:
                task = self.reorder_buffer.popleft()
                yield task
                try:
                    self.next_id = self.get_sequence_id(task) + 1
                except:
                    pass
            else:
                break

    def run(self, task) -> Iterable[Task]:
        assert isinstance(task, Task)
        self.logger.debug('Enqueue: %s', task)
        self.push(task)
        for task in self.iter_pop():
            try:
                if not task.success:
                    self.logger.warn('Bypassing: %s', task)
                    yield task
                    continue
                self.logger.debug('Processing: %s', task)
                result = self.process(task)
                if result is None:
                    continue
                elif isinstance(result, Task):
                    yield result
                else:
                    for res in result:
                        yield res
                self.logger.debug('Processed: %s', task)
            except Exception as e:
                task.fail()
                self.logger.exception('Failed: %s', task)
                if Options.raise_exception:
                    raise e
                else:
                    yield task



import os
from copy import deepcopy
from typing import Any, Dict, List, Optional

import GPUtil


class Resources(object):

    '''
    Resource management and allocation.
    '''

    def __init__(self, resources: Dict[str, List[Any]] = None, **scan_args):
        if resources is not None:
            self.resources = resources
        else:
            self.resources = self.scan(**scan_args)

    def __add__(self, other):
        new_resources = deepcopy(self.resources)
        for key, value in other.resources.items():
            new_resources.setdefault(key, []).extend(value)
        return self.__class__(new_resources)

    def get(self, name: str, limit: Optional[int] = None):
        return self.resources.get(name, [])[:limit]

    def scan(self, gpu_load: float = 0.1, gpu_memory: float = 0.1):
        '''
        Initial global scan for a system.
        '''
        resources = {'cpu': [*range(os.cpu_count())]}
        gpus = GPUtil.getAvailable(
            limit=100, maxLoad=gpu_load, maxMemory=gpu_memory)
        if len(gpus) > 0:
            resources['gpu'] = gpus
        return resources

    def split(self, num_split: int):
        '''
        Even split between pipelines in a system, drop remainders.
        '''
        results = [{} for _ in range(num_split)]
        for key, value in self.resources.items():
            split_size = len(value) // num_split
            for split_i in range(num_split):
                results[split_i][key] = value[
                    split_i * split_size: (split_i + 1) * split_size]
        results = [Resources(r) for r in results]
        return results

    def select(self, **select_dict):
        '''
        Unbalanced select for stages in a pipeline.
        select_dict value: True for all, False for None, 
            (start, end) in int/float/None for slice
        '''
        selected_resources = {}
        for key, value in select_dict.items():
            if not key in self.resources:
                continue
            resource = self.resources[key]
            if value is True:
                selected_resources[key] = resource
            elif value is False:
                continue
            else:
                start, end = self._index_or_frac(value, len(resource))
                selected_resources[key] = resource[start:end]
        return Resources(selected_resources)

    @staticmethod
    def _index_or_frac(values, length):
        is_float = any([isinstance(value, float) for value in values])
        if not is_float:
            return values
        return [int(round(value * length)) if value is not None else None
                for value in values]

    def __str__(self):
        return ', '.join(['%s=%s' % (k, v) for k, v in self.resources.items()])

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, str(self))
